from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Literal
from urllib.parse import urlparse

import httpx

from mokr.constants import HTTP_METHODS
from mokr.network.request import Request
from mokr.network.response import Response

if TYPE_CHECKING:
    from mokr.browser.page import Page


class HttpDomain():
    def __init__(
        self,
        page: Page,
        sync_cookies: Literal["both", "http", "page", "none"] = "both",
        **httpx_client_kwargs,
    ) -> None:
        """
        Client to handle sending requests outside of the browser, while syncing
        state between itself and the parent `mokr.browser.Page`.

        This client uses `httpx` so requests can be made with HTTP2.

        Args:
            page (Page): Parent `mokr.browser.Page` that this will sync with.
            sync_cookies (Literal["both", "http", "page", "none"], optional):
                Whether to sync cookies between this client and the parent
                `mokr.browser.Page` before/after each request and response.
                Defaults to "both".
        """
        self._page = page
        self._cookie_sync_directive = sync_cookies
        self._build_proxy()
        kwargs = dict(
            http2=True,
            proxies=self._proxies,
            verify=not self._page._ignore_https_errors,
            timeout=self._page._default_navigation_timeout,
            follow_redirects=True,
        )
        if httpx_client_kwargs:
            kwargs.update(**httpx_client_kwargs)
        headers = {"User-Agent": self._page.user_agent}
        if "headers" in kwargs:
            kwargs["headers"].update(headers)
        else:
            kwargs["headers"] = headers
        # Cannot use the AsyncClient due to SOCKS support issue:
        # https://github.com/encode/httpx/discussions/2305
        # For now, throwing methods into threads via asyncio.
        self._client = httpx.Client(**kwargs)

    def _build_proxy(self) -> None:
        creds = self._page._proxy_credentials.copy()
        if not creds:
            self._proxies = None
            return
        if not creds.get("username"):
            proxy = creds["proxy"]
        else:
            proxy_parts = urlparse(creds["proxy"])
            scheme = proxy_parts.scheme
            host = proxy_parts.hostname
            port = proxy_parts.port
            username = creds.get("username", "")
            auth_block = ""
            if username:
                auth_block = f'{username}:{creds.get("password", "")}'
            proxy = f"{scheme}://{auth_block}@{host}:{port}"
        self._proxies = {"http://": proxy, "https://": proxy}
        if proxy.startswith("socks"):
            self._proxies["socks5://"] = proxy

    async def _sync_cookies_from_browser(self) -> None:
        browser_cookies = await self._page.cookies()
        for cookie in browser_cookies:
            self._client.cookies.set(
                cookie["name"],
                cookie["value"],
                cookie.get("domain", ""),
                cookie.get("path", "/"),
            )

    async def _sync_cookies_to_browser(self) -> None:
        cookies_list = []
        for data in self._client.cookies.jar._cookies.values():
            for cookie_map in data.values():
                for cookie in cookie_map.values():
                    cookie_dump = {
                        "name": cookie.name,
                        "value": cookie.value,
                        "domain": cookie.domain,
                        "path": cookie.path,
                        "expires": cookie.expires,
                        "secure": cookie.secure,
                    }
                    cookie_dump = {k: v for k, v in cookie_dump.items() if v}
                    if cookie_dump:
                        cookies_list.append(cookie_dump)
        if cookies_list:
            await self._page.set_cookies(cookies_list)

    async def _sync_cookies(self):
        if not self._cookie_sync_directive:
            return
        if self._cookie_sync_directive in ("both", "http"):
            await self._sync_cookies_from_browser()
        if self._cookie_sync_directive in ("both", "page"):
            await self._sync_cookies_to_browser()

    async def _http_to_mokr(self, response: httpx.Response) -> Request:
        request = response.request
        data = await asyncio.to_thread(request.read)
        payload = {
            "method": request.method,
            "postData": data.decode(),
            "headers": dict(request.headers),
        }
        mokr_request = Request(
            self._page,
            self._page._client,
            "httpxrequest",
            None,
            False,
            False,
            str(request.url),
            "Other",
            payload,
            None,
            [],
            [],
            httpx_request=request,
        )
        mokr_response = Response(
            self._page._client,
            mokr_request,
            response.status_code,
            response.headers,
            False,
            False,
            False,
            httpx_response=response,
        )
        mokr_response._request = mokr_request
        mokr_request._response = mokr_response
        return mokr_request

    async def _transform_http_objs(self, response: httpx.Response) -> Response:
        http_redirect_chain = [response]
        if response.history:
            http_redirect_chain = http_redirect_chain + response.history.copy()
        requests = [(await self._http_to_mokr(r)) for r in http_redirect_chain]
        if len(requests) > 1:
            # The httpx history is reverse to what our redirect_chain is.
            requests.reverse()
            for request in requests:
                request._redirect_chain = requests
        return requests[-1].response

    async def send(
        self,
        url: str,
        method: HTTP_METHODS = "GET",
        params: dict | None = None,
        headers: dict | None = None,
        data: dict | None = None,
        json: dict | None = None,
        **httpx_request_kwargs,
    ) -> Response:
        """
        Send an HTTP request.

        Args:
            url (str): The url to request.
            method (Literal[str], optional): HTTP method. Defaults to "GET".
            params (dict | None, optional): Parameters to encode in the URL
                before sending. Defaults to None (omitted).
            headers (dict | None, optional): Additional headers to send with
                this request beyond what was set when the client was created.
                Defaults to None (omitted).
            data (dict | None, optional): Data payload to deliver with request
                (form-encoded data). Only for "PUT", "POST", and "PATCH".
                Defaults to None (omitted).
            json (dict | None, optional): Data payload to deliver with request
                (JSON-encoded data). Only for "PUT", "POST", and "PATCH".
                Defaults to None (omitted).

        Raises:
            ValueError: Raised if invalid HTTP method given or `data` or `json`
                arguments given with incompatible methods.

        Returns:
            Response: A `mokr.network.Response` created from the
                `httpx.Response`. All redirects will also be created, and
                original requests and responses will be accessible from the
                `mokr.network` objects.
        """
        method = method.upper()
        methods = ("GET", "POST", "OPTIONS", "HEAD", "PUT", "PATCH", "DELETE")
        data_methods = ("POST", "PUT", "PATCH")
        if method not in methods:
            raise ValueError(f"Invalid method: {method}, must be of: {methods}")
        if (data or json) and method not in data_methods:
            raise ValueError(f"Can only send data on methods: {data_methods}")
        kwargs = dict(params=params, headers=headers, data=data, json=json)
        if httpx_request_kwargs:
            kwargs.update(**httpx_request_kwargs)
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        await self._sync_cookies()
        response = await asyncio.to_thread(
            getattr(self._client, method.lower()),
            url,
            **kwargs,
        )
        await self._sync_cookies()
        return await self._transform_http_objs(response)



from __future__ import annotations

import asyncio
import concurrent.futures
import json
import uuid
from typing import TYPE_CHECKING, Any, Literal

from pyee import EventEmitter

from mokr.constants import (
    HTTP_CACHE_TYPES,
    HTTP_METHODS,
    FETCH_DOMAIN_REQUEST_SENT,
    METHOD_FETCH_REQUEST,
    NETWORK_REQUEST,
    NETWORK_RESPONSE,
    REFERRER_POLICIES,
)
from mokr.exceptions import MokrTimeoutError, NetworkError, PageError
from mokr.execution.handle.javascript import JavascriptHandle
from mokr.network.request import Request
from mokr.network.response import Response
from mokr.waiters import EventWaiter

if TYPE_CHECKING:
    from mokr.browser.page import Page


class _FirefoxResponseReady(Exception):
    """Used to cancel timeout and transfer response data."""
    def __init__(self, data: dict) -> None:
        self.data = data


class FetchDomain(EventEmitter):
    def __init__(
        self,
        page: Page,
    ) -> None:
        """
        Client to handle sending manual fetch requests.

        Note that sending fetch requests from this class will temporarily
        enable request interception globally. However, these fetch requests
        won't be caught by the request interception chain, however "response"
        and "requestfinished" events will still be caught.

        For Firefox this is not applicable. Fetch requests in Firefox
        are built manually from the response data. Unlike standard Firefox
        request responses, the response body is available for `FetchDomain`
        responses.

        Chrome requests take advantage of the existing fetch system in its
        `mokr.network.NetworkManager` while Firefox cannot; Chrome could
        do the same as Firefox does here and build the objects directly but
        some features would be lost like redirect chain.

        Args:
            page (Page): Parent `mokr.browser.Page` this was spawned from.
        """
        super().__init__()
        self._page = page
        self._requests_in_flight = set()
        self._request_id_to_response: dict[str, Request] = {}
        self._is_firefox = self._page._is_firefox(mute=True)
        self._page.on(NETWORK_RESPONSE, self._response_intercept)

    async def _request_has_response(self, request: Request) -> Response:
        if not request._mokr_request_uuid:
            return
        while True:
            done = False
            if not request.redirect_chain and request.response:
                done = True
            elif request.redirect_chain:
                for _request in request.redirect_chain:
                    response = _request.response
                    # If redirect, wait for the final response.
                    if response and not 299 < response.status < 400:
                        done = True
                    else:
                        done = False
            if not done:
                # Yield while waiting for response to be recieved.
                await asyncio.sleep(0)
            else:
                break
        return True

    async def _register_requests(self, request: Request) -> Request:
        if not request._mokr_request_uuid:
            return request
        else:
            self.emit(FETCH_DOMAIN_REQUEST_SENT, request)
            self._request_id_to_response[request._request_id] = None
            # Bypass the rest of the existing request intercept chain.
            await request.release()

    async def _evaluate_handle(
        self,
        page_function: str,
        *args: Any,
        eval_script_url_suffix: str | None = None,
    ) -> JavascriptHandle:
        frame = self._page._ensure_frame()
        context = await frame.execution_context()
        if not context:
            raise PageError('No context attached to frame.')
        try:
            return await context.evaluate_handle(
                page_function,
                *args,
                eval_script_url_suffix=eval_script_url_suffix,
            )
        except NetworkError as error:
            # Ignore, we likely intentionally destroyed it due to timeout.
            if "Execution context destroyed" in str(error):
                pass
            else:
                raise

    async def _fetch(self, request_uuid: str, url: str, params: dict) -> None:
        self._requests_in_flight.add(request_uuid)
        # The uuid is tacked onto the script name and extracted in interception.
        response = await self._evaluate_handle(
            METHOD_FETCH_REQUEST,
            url,
            params,
            eval_script_url_suffix=request_uuid,
        )
        try:
            response_data = json.loads(response._remote_object["value"])
            # Use this to cancel timer in Firefox as predicate will be useless
            # since the stack in browser won't share the javascript url with
            # the uuid in it.
            if self._is_firefox:
                raise _FirefoxResponseReady(response_data)
        except (json.JSONDecodeError, TypeError, KeyError):
            return None

    def _response_intercept(self, response: Response) -> None:
        request = response._request
        if not request._mokr_request_uuid:
            return
        self._request_id_to_response[response._request._request_id] = response

    def _select_response_chrome(self, request_uuid: str) -> Response:
        responses = {
            request_id: response for request_id, response in
            self._request_id_to_response.items()
            if response and response._request._mokr_request_uuid == request_uuid
        }
        # This should never happen.
        if not responses:
            raise NetworkError(
                "Fetch request failed, no response could be found for request."
                " The destination URL may have a CORS policy enabled that this"
                " page does not have access to."
            )
        request_id = next(iter(responses))
        return self._request_id_to_response.pop(request_id)

    def _make_response_firefox(self, response_data: dict) -> Response:
        request = response_data["request"]
        payload = {
            "method": request["method"],
            "postData": request.get("body", ""),
            "headers": request.get("headers", {}),  # Will be incomplete.
        }
        mokr_request = Request(
            self._page,
            self._page._client,
            "firefoxfetch",
            None,
            False,
            False,
            response_data["url"],
            response_data["type"],
            payload,
            None,
            [],
            [],
        )
        mokr_response = Response(
            self._page._client,
            mokr_request,
            response_data["status"],
            response_data["headers"],
            False,
            False,
            True,
        )
        mokr_response._request = mokr_request
        mokr_response._content = response_data["body"]
        mokr_request._response = mokr_response
        return mokr_response

    def _select_response(self, request_uuid: str, data: dict) -> Response:
        if self._is_firefox:
            return self._make_response_firefox(data)
        else:
            return self._select_response_chrome(request_uuid)

    def _check_for_response_data(
        self,
        task: asyncio.Task,
        params: dict,
    ) -> dict | None:
        # Check if Firefox fetch resolved.
        response_data = None
        error = task.exception()
        if error and isinstance(error, _FirefoxResponseReady):
            response_data = error.data
            # Pass params to build request from, too.
            response_data["request"] = params
        elif error:
            # This should never happen here...
            raise MokrTimeoutError("Timeout waiting for fetch to resolve.")
        return response_data

    async def fetch(
        self,
        url: str | None = None,
        timeout: int = None,
        request: Request | None = None,
        body: str | None = None,
        browsing_topics: bool | None = None,
        cache: HTTP_CACHE_TYPES | None = None,
        credentials: Literal["omit", "same-origin", "include"] | None = None,
        headers: dict | None = None,
        method: HTTP_METHODS = "GET",
        mode: Literal["cors", "no-cors", "same-origin"] = None,
        priority: Literal["high", "low", "auto"] | None = None,
        redirect: Literal["follow", "error", "manual"] | None = None,
        referrer: str | None = None,
        referrer_policy: REFERRER_POLICIES | None = None,
    ) -> Response:
        """
        Send a fetch request.

        Note that fetch is sensitive to the current site's CORS settings.
        Additionally, trying to send a fetch request before the DOM loads can
        hang forever.

        Args:
            url (str | None, optional): The url to request.
                Ignored if `request` is given.
                Defaults to None.
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to None.
            request (Request | None, optional): A request object to pull `url`,
                `method`, and `headers` from. Defaults to None.
            body (str | None, optional): Body to add to the request.
                Defaults to None.
            browsing_topics (bool | None, optional): If True, selected topics
                will be sent in a Sec-Browsing-Topics header with the associated
                request.
                Defaults to None.
            cache (HTTP_CACHE_TYPES): How the request should interact with the
                HTTP cache. One of: "default", "no-store", "reload", "no-cache",
                "force-cache", or "only-if-cache".
                Defaults to None.
            credentials (Literal[str] | None, optional): How to handle
                credentials. One of "omit", "same-origin", or "include".
                Defaults to None.
            headers (dict | None, optional): Headers to pass with the request.
                Defaults to None.
            method (HTTP_METHODS, optional): HTTP method. Defaults to "GET".
            mode (Literal["cors", "no-cors", "same-origin"], optional):
                CORS mode to use. Defaults to None.
            priority (Literal["high", "low", "auto"] | None, optional):
                Priority of request relative to others. Defaults to None.
            redirect (Literal["follow", "error", "manual"] | None, optional):
                How to handle redirects. Defaults to None ("follow").
            referrer (str | None, optional): Request referrer. Defaults to None.
            referrer_policy (REFERRER_POLICIES): The referrer policy to use.
                One of "no-referrer", "no-referrer-when-downgrade",
                "same-origin", "origin", "strict-origin",
                "origin-when-cross-origin", "strict-origin-when-cross-origin",
                or "unsafe-url".
                Defaults to None ("strict-origin-when-cross-origin").

        Raises:
            ValueError: Raised if neither `url` nor `request` given.
            NetworkError: Raised if no response found when request resolves.

        Returns:
            Response: Last response recieved after all redirects, if any.
        """
        request_uuid = uuid.uuid4().hex.upper()
        interception_enabled = (
            self._page._network_manager._protocol_request_interception_enabled
        )
        # Must enable request interception, if not already enabled.
        if not interception_enabled and not self._is_firefox:
            await self._page.set_request_interception_enabled(True)
        # Done here instead of in constructor, should be 0th priority.
        self._page.on(NETWORK_REQUEST, self._register_requests)
        if request:
            url = request.url
            headers = request.headers
            method = request.method
        elif not url:
            raise ValueError("Must provide URL or request.")
        params = {
            name: value for name, value in {
                "body": body,
                "browsingTopics": browsing_topics,
                "cache": cache,
                "credentials": credentials,
                "headers": headers,
                "method": method,
                "mode": mode,
                "priority": priority,
                "redirect": redirect,
                "referrer": referrer,
                "referrerPolicy": referrer_policy,
            }.items() if value is not None
        }
        waiter = EventWaiter(
            self,
            FETCH_DOMAIN_REQUEST_SENT,
            self._request_has_response,
            timeout if timeout else self._page._default_navigation_timeout,
            self._page._client._loop,
        )
        # Setup waiter and make request.
        promise = asyncio.ensure_future(waiter.wait())
        done, _ = await asyncio.wait(
            [
                promise,
                self._fetch(request_uuid, url, params),
            ],
            return_when=concurrent.futures.FIRST_EXCEPTION,
        )
        if promise.done() and promise.exception():
            raise MokrTimeoutError("Timeout waiting for fetch to resolve.")
        response_data = self._check_for_response_data(done.pop(), params)
        self._requests_in_flight.remove(request_uuid)
        # Remove intercept callback if no more FetchDomain requests in flight.
        if not self._requests_in_flight:
            self._page._interception_callback_chain.remove(
                self._register_requests
            )
            if not interception_enabled and not self._is_firefox:
                await self._page.set_request_interception_enabled(False)
        response = self._select_response(request_uuid, response_data)
        return response



from __future__ import annotations

from mokr.connection.connection import DevtoolsConnection
from mokr.constants import EMULATION_ENABLE_TOUCH, EMULATION_OVERRIDE_METRICS


class ViewportManager():
    def __init__(self, client: DevtoolsConnection) -> None:
        """
        Handler for adjusting the browser viewport. Can adjust size and
        enable functionality such as touchmode.

        Args:
            client (DevtoolsConnection): `mokr.network.DevtoolsConnection`
                object, shared from the parent `mokr.browser.Page`.
        """
        self._client = client
        self._emulating_mobile = False
        self._has_touch = False

    async def emulate_viewport(self, viewport: dict[str, bool | int]) -> bool:
        """
        Run viewport adjustments based off given `viewport` parameters.
        Not all viewport options are considered, only: "isMobile", "width",
        "height", "deviceScaleFactor", "isLandscape", and "hasTouch".

        Args:
            viewport (dict[str, bool  |  int]): The parameters to adjust
                the viewport to.

        Raises:
            TypeError: Raised if "width" or "height" in `viewport` are not
                positive integers.

        Returns:
            bool: True to indicate if a reload of the page is needed to affect
                the changes sent. False if not.
        """
        options = {}
        mobile = viewport.get('isMobile', False)
        options['mobile'] = mobile
        for viewport_axis in ("width", "height"):
            axis_value = viewport.get(viewport_axis)
            if axis_value:
                if not isinstance(axis_value, int) or axis_value < 0:
                    raise TypeError(
                        'Viewport dimensions must be positive integers'
                        f', got: {axis_value} ({type(axis_value)})'
                    )
                options[viewport_axis] = axis_value
        options['deviceScaleFactor'] = viewport.get('deviceScaleFactor', 1)
        if viewport.get('isLandscape'):
            options['screenOrientation'] = {
                'angle': 90,
                'type': 'landscapePrimary',
            }
        else:
            options['screenOrientation'] = {
                'angle': 0,
                'type': 'portraitPrimary',
            }
        has_touch = viewport.get('hasTouch', False)
        await self._client.send(EMULATION_OVERRIDE_METRICS, options)
        await self._client.send(
            EMULATION_ENABLE_TOUCH,
            {
                'enabled': has_touch,
                'configuration': 'mobile' if mobile else 'desktop',
            }
        )
        reload_needed = (
            self._emulating_mobile != mobile or self._has_touch != has_touch
        )
        self._emulating_mobile = mobile
        self._has_touch = has_touch
        return reload_needed



from mokr.network.request import Request


class NetworkEventManager():
    def __init__(self) -> None:
        """
        This class manages a collection of dictionaries used by
        `mokr.network.NetworkManager` to track active event status as events can
        be received in multiple different orders.

        See `mokr.network.NetworkManager` for more information.
        """
        # NetworkRequestId to Network.RequestWillBeSentEvent
        self.request_will_be_sent_map = {}
        # NetworkRequestId to Fetch.RequestPausedEvent
        self.request_paused_map = {}
        # NetworkRequestId to CdpHTTPRequest
        self.http_requests_map = {}
        # NetworkRequestId to Network.ResponseReceivedExtraInfoEvent
        self.response_received_extra_info_map = {}
        # NetworkRequestId to RedirectInfoList
        self.queued_redirect_info_map = {}
        # NetworkRequestId to QueuedEventGroup
        self.queued_event_group_map = {}
        # The response_received_extra_info_map, queued_redirect_info_map,
        # and queued_event_group_map maps are used to reconcile
        # "Network.ResponseReceivedExtraInfo" events with their corresponding
        # request.
        # Each response and redirect response gets an "ExtraInfo" event, but we
        # don't know which will come first.
        # To resolve this, we store a Response or an ExtraInfo for each
        # response, and the `NetworkManager` emits the event when we have both.
        # As arrays they model the chain of events and handle redirects.

    def forget(self, network_id: str) -> None:
        """
        Remove data for the given `network_id` from all mappings.

        Args:
            network_id (str): Request network identifier.
        """
        for request_map in [
            self.request_will_be_sent_map,
            self.request_paused_map,
            self.queued_event_group_map,
            self.queued_redirect_info_map,
            self.response_received_extra_info_map,
        ]:
            try:
                request_map.pop(network_id)
            except KeyError:
                pass

    def response_extra_info(self, network_id: str) -> list[dict]:
        """
        Get stored response extraInfo for given `network_id`.

        Args:
            network_id (str): Request network identifier.

        Returns:
            list[dict]: List of extraInfo events as dictionaries.
        """
        if not self.response_received_extra_info_map.get(network_id):
            self.response_received_extra_info_map[network_id] = []
        return self.response_received_extra_info_map[network_id]

    def queued_redirect_info(self, fetch_id: str) -> list[dict]:
        """
        Get stored redirect info for given `fetch_id`.

        Args:
            fetch_id (str): Request fetch identifier.

        Returns:
            list[dict]: List of redirect info as dictionaries.
        """
        if not self.queued_redirect_info_map.get(fetch_id):
            self.queued_redirect_info_map[fetch_id] = []
        return self.queued_redirect_info_map[fetch_id]

    def queue_redirect_info(
        self,
        fetch_id: str,
        redirect_info: list[dict],
    ) -> None:
        """
        Store redirect info under a given `fetch_id`.

        Args:
            fetch_id (str): Request fetch identifier.
            redirect_info (list[dict]): Redirect info data.
        """
        self.queued_redirect_info_map[fetch_id] = redirect_info

    def take_queued_redirect_info(self, fetch_id: str) -> list[dict]:
        """
        Remove and return redirect info stored under given `fetch_id`.

        Args:
            fetch_id (str): Request fetch identifier.

        Returns:
            list[dict]: Redirect info data.
        """
        if self.queued_redirect_info(fetch_id):
            return self.queued_redirect_info_map.pop(fetch_id)
        else:
            return []

    def get_in_flight_requests_count(self) -> int:
        """
        Get the number of active requests (not resolved).

        Returns:
            int: Number of active requests.
        """
        in_flight_requests_counter = 0
        for request in self.http_requests_map.values():
            if not request.response():
                in_flight_requests_counter += 1
        return in_flight_requests_counter

    def store_request_will_be_sent(self, network_id: str, event: dict) -> None:
        """
        Store a requestWillBeSent event under given `network_id`.

        Args:
            network_id (str): Request network identifier.
            event (dict): The requestWillBeSent event received.
        """
        self.request_will_be_sent_map[network_id] = event

    def get_request_will_be_sent(self, network_id: str) -> dict:
        """
        Get requestWillBeSent event stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.

        Returns:
            dict: The stored requestWillBeSent under given `network_id`.
        """
        return self.request_will_be_sent_map.get(network_id)

    def forget_request_will_be_sent(self, network_id: str) -> None:
        """
        Remove requestWillBeSent stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.
        """
        self.request_will_be_sent_map.pop(network_id)

    def get_request_paused(self, network_id: str) -> dict:
        """
        Get requestPaused event stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.

        Returns:
            dict: The stored requestPaused under given `network_id`.
        """
        return self.request_paused_map.get(network_id)

    def forget_request_paused(self, network_id: str) -> None:
        """
        Remove requestPaused stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.
        """
        self.request_paused_map.pop(network_id)

    def store_request_paused(self, network_id: str, event: dict) -> None:
        """
        Store a requestPaused event under given `network_id`.

        Args:
            network_id (str): Request network identifier.
            event (dict): The requestPaused event received.
        """
        self.request_paused_map[network_id] = event

    def get_request(self, network_id: str) -> Request:
        """
        Get `mokr.network.Request` stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.

        Returns:
            dict: The stored `mokr.network.Request` under given `network_id`.
        """
        return self.http_requests_map.get(network_id)

    def store_request(self, network_id: str, request: Request) -> None:
        """
        Store a `mokr.network.Request` under given `network_id`.

        Args:
            network_id (str): Request network identifier.
            request (Request): The `mokr.network.Request` received.
        """
        self.http_requests_map[network_id] = request

    def forget_request(self, network_id: str) -> None:
        """
        Remove `mokr.network.Request` stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.
        """
        self.http_requests_map.pop(network_id)

    def get_queued_event_group(self, network_id: str) -> dict:
        """
        Get responseReceivedEvent event stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.

        Returns:
            dict: The stored responseReceivedEvent event under given
                `network_id`.
        """
        return self.queued_event_group_map.get(network_id)

    def queue_event_group(self, network_id: str, event: dict) -> None:
        """
        Store a responseReceivedEvent event under given `network_id`.

        Args:
            network_id (str): Request network identifier.
            event (dict): The responseReceivedEvent received.
        """
        self.queued_event_group_map[network_id] = event

    def forget_queued_event_group(self, network_id: str) -> None:
        """
        Remove responseReceivedEvent event stored under given `network_id`.

        Args:
            network_id (str): Request network identifier.
        """
        self.queued_event_group_map.pop(network_id)



from __future__ import annotations

from subprocess import Popen
from typing import Any, Awaitable, Callable, Literal

from pyee import EventEmitter

from mokr.browser.context import BrowserContext
from mokr.browser.page import Page
from mokr.browser.target import Target
from mokr.connection import Connection
from mokr.constants import (
    BROWSER_GET_VERSION,
    DISCONNECTED,
    TARGET_CHANGED,
    TARGET_CREATE_BROWSER_CONTEXT,
    TARGET_CREATE_TARGET,
    TARGET_CREATED,
    TARGET_DESTROYED,
    TARGET_DISPOSE_BROWSER_CONTEXT,
    TARGET_INFO_CHANGED,
    TARGET_SET_DISCOVER_TARGETS,
    TARGET_TARGET_CREATED,
    TARGET_TARGET_DESTROYED,
)
from mokr.exceptions import BrowserError


class Browser(EventEmitter):
    def __init__(
        self,
        browser_type: Literal["chrome", "firefox"],
        connection: Connection,
        context_ids: list[str],
        ignore_https_errors: bool,
        default_viewport: dict | None,
        process: Popen | None = None,
        close_callback: Callable | None = None,
        proxy_credentials: dict | None = None,
        default_user_agent: str | None = None,
        **kwargs: Any,
    ) -> None:
        """
        This class is created upon connect to a browser. It is essentially a
        container for the individual pages and browser contexts.

        Args:
            browser_type (Literal["chrome", "firefox"]): The type of browser.
            connection (`mokr.network.Connection`): Websocket connection.
            context_ids (list[str]): Browser context identifiers.
            ignore_https_errors (bool): Ignore site security errors.
            default_viewport (dict | None): Default viewport configuration.
            process (Popen | None, optional): Local browser process, if any.
                Defaults to None.
            close_callback (Callable | None, optional): Callback to run on
                close. Defaults to None.
            proxy_credentials: (dict | None, optional): Dictionary with proxy
                credentials keyed as "username" and "password". Credentials
                should be for the proxy the browser process is bound to.
            default_user_agent (str, optional): Default user agent to use on
                all new pages.
        """
        super().__init__()
        self._browser_type = browser_type
        self._ignore_https_errors = ignore_https_errors
        self._default_viewport = default_viewport
        self._default_user_agent = default_user_agent
        self._process = process
        self._screenshot_task_queue: list = []
        self._connection = connection
        self._proxy_credentials = proxy_credentials
        self._version = None
        loop = self._connection._loop
        if close_callback:
            self._close_callback = close_callback
        else:
            self._close_callback = self._dummy_callback
        self._default_context = BrowserContext(self, None)
        self._contexts: dict[str, BrowserContext] = dict()
        for context_id in context_ids:
            self._contexts[context_id] = BrowserContext(self, context_id)
        self._targets: dict[str, Target] = dict()
        self._connection._set_closed_callback(
            lambda: self.emit(DISCONNECTED)
        )
        self._connection.on(
            TARGET_TARGET_CREATED,
            lambda event: loop.create_task(self._target_created(event)),
        )
        self._connection.on(
            TARGET_TARGET_DESTROYED,
            lambda event: loop.create_task(self._target_destroyed(event)),
        )
        self._connection.on(
            TARGET_INFO_CHANGED,
            lambda event: loop.create_task(self._target_info_changed(event)),
        )

    @property
    def kind(self) -> str:
        """One of "chrome" or "firefox"."""
        return self._browser_type

    @property
    def process(self) -> Popen | None:
        """The local browser process. If created via `mokr.connect`, will
        return None.
        """
        return self._process

    @property
    def version(self) -> str:
        """Get browser version (product from browser full version info.)"""
        return self._version['product'] if self._version else ''

    @property
    def user_agent(self) -> str:
        """
        Get the user agent the browser was spawned with or that is set as the
        default to override with. This can be overidden later again
        with `mokr.browser.Page.set_user_agent`.
        """
        default = self._version.get("userAgent", '') if self._version else ''
        return self._default_user_agent if self._default_user_agent else default

    @property
    def browser_contexts(self) -> list[BrowserContext]:
        """
        A list of all `mokr.browser.BrowserContext` instances attached to this
        browser. By default, this will be a single context.
        """
        return [self._default_context] + [
            context for context in self._contexts.values()
        ]

    @property
    def ws_endpoint(self) -> str:
        """
        The websocket URL that this `Browser`'s `mokr.connection.Connection`
        object is using.
        """
        return self._connection.url

    async def _get_version(self) -> Awaitable:
        return await self._connection.send(BROWSER_GET_VERSION)

    def _dummy_callback(self) -> Awaitable[None]:
        fut = self._connection._loop.create_future()
        fut.set_result(None)
        return fut

    def _emit_cascade(
        self,
        event: str,
        context: BrowserContext,
        *args,
        **kwargs,
    ) -> bool:
        # Emit the same event on this object and the given context.
        obj_result = self.emit(event, *args, **kwargs)
        context_result = context.emit(event, *args, **kwargs)
        return obj_result and context_result

    async def _dispose_context(self, context_id: str) -> None:
        await self._connection.send(
            TARGET_DISPOSE_BROWSER_CONTEXT,
            {'browserContextId': context_id},
        )
        self._contexts.pop(context_id, None)

    async def _target_created(self, event: dict) -> None:
        target_info = event['targetInfo']
        browser_context_id = target_info.get('browserContextId')
        if browser_context_id and browser_context_id in self._contexts:
            context = self._contexts[browser_context_id]
        else:
            context = self._default_context
        # Indicate if overriden to avoid needless override with the default.
        user_agent_data = {
            "overridden": True if self._default_user_agent else False,
            "user_agent": self.user_agent,
        }
        target = Target(
            self,
            target_info,
            context,
            lambda: self._connection.create_session(target_info),
            self._ignore_https_errors,
            self._default_viewport,
            self._screenshot_task_queue,
            self._connection._loop,
            self._proxy_credentials,
            user_agent_data,
        )
        if target_info['targetId'] in self._targets:
            raise BrowserError('Target should not exist before create.')
        self._targets[target_info['targetId']] = target
        if await target._initialized_promise:
            self._emit_cascade(TARGET_CREATED, context, target)

    async def _target_destroyed(self, event: dict) -> None:
        target = self._targets.pop(event['targetId'])
        target._closed_callback()
        if await target._initialized_promise:
            self._emit_cascade(
                TARGET_DESTROYED,
                target.browser_context,
                target,
            )
        target._initialized_callback(False)

    async def _target_info_changed(self, event: dict) -> None:
        target = self._targets.get(event['targetInfo']['targetId'])
        if not target:
            raise BrowserError('Target should exist before info changed.')
        previous_url = target.url
        was_initialized = target._is_initialized
        target._target_info_changed(event['targetInfo'])
        if was_initialized and previous_url != target.url:
            self._emit_cascade(
                TARGET_CHANGED,
                target.browser_context,
                target,
            )

    async def _create_page_in_context(self, context_id: str | None) -> Page:
        options = {'url': 'about:blank'}
        if context_id:
            options['browserContextId'] = context_id
        response = await self._connection.send(TARGET_CREATE_TARGET, options)
        target_id = response.get('targetId')
        target = self._targets.get(target_id)
        if target is None or not await target._initialized_promise:
            raise BrowserError('Failed to create target for page.')
        page = await target.page()
        if page is None:
            raise BrowserError('Failed to create page.')
        return page

    async def first_page(self) -> Page | None:
        """
        Return the first page in the default context's `BrowserContext.pages`.
        If no pages active, returns None.

        Returns:
            Page | None: First active page in the default context, if any.
        """
        return await self._default_context.first_page()

    async def ready(self) -> Browser:
        """
        Enable target discovery in the remote connection.

        Returns:
            Browser: This `Browser` class.
        """
        self._version = await self._get_version()
        await self._connection.send(
            TARGET_SET_DISCOVER_TARGETS,
            {'discover': True},
        )
        return self

    async def create_incognito_browser_context(self) -> BrowserContext:
        """
        Create a new browser context. This is akin to spawning an incognito
        browser window, it will not share cookies or storage with pages
        in other contexts. To do so, use `Browser.new_page` instead.

        Example::

            browser = await launch().launch()
            # Navigate and login or perform another storage-accessed action.
            context = await browser.create_incognito_browser_context()
            page = await context.first_page()
            # Navigate to the same site. Session isn't shared!
            ...

        Returns:
            BrowserContext: A new `mokr.browser.BrowserContext`.
        """
        obj = await self._connection.send(TARGET_CREATE_BROWSER_CONTEXT)
        browser_context_id = obj['browserContextId']
        context = BrowserContext(self, browser_context_id)
        self._contexts[browser_context_id] = context
        return context

    async def new_page(self) -> Page:
        """
        Spawn a new page within the default context.

        Returns:
            Page: A new `mokr.browser.Page` at "about:blank".
        """
        return await self._default_context.new_page()

    def targets(self) -> list[Target]:
        """
        A list of all `mokr.browser.Target`s in all contexts attached to this
        `Browser` object.

        Returns:
            list[Target]: All initialised targets within all contexts in
                this browser.
        """
        return [
            target for target in self._targets.values()
            if target._is_initialized
        ]

    async def pages(self) -> list[Page]:
        """
        A list of all `mokr.browser.Page`s in all contexts attached to this
        `Browser` object.

        Returns:
            list[Page]: All pages within all contexts in this browser.
        """
        return [
            page for page_list in [
                await context.pages() for context in self.browser_contexts
            ]
            for page in page_list
        ]

    async def close(self) -> None:
        """Run the `close_callback` given during initialisation."""
        await self._close_callback()

    async def disconnect(self) -> None:
        """
        Disconnect the `Browser`'s `Connection` object's websocket connection
        and fail any `Browser.targets` that haven't finished initialising.
        """
        await self._connection.dispose()
        for target in self._targets.values():
            if not target._is_initialized:
                target._initialized_callback(False)



from pathlib import Path

from mokr.constants import javascript


root = Path(javascript.__file__).parent

METHOD_FOCUS_ELEMENT = 'element => element.focus()'

METHOD_GET_PROPERTY = (root / "get_property.js").read_text()
METHOD_SCROLL_INTO_VIEW = (root / "scroll_into_view.js").read_text()
METHOD_EVAL_XPATH = (root / "evaluate_xpath.js").read_text()
METHOD_ELEMENT_IN_VIEW = (root / "element_in_view.js").read_text()
METHOD_WAIT_FOR_PREDICATE_PAGE = (root / "wait_for_page.js").read_text()
METHOD_GET_CONTENT = (root / "get_content.js").read_text()
METHOD_SET_CONTENT = (root / "set_content.js").read_text()
METHOD_EMBED_JAVASCRIPT_BY_URL = (root / "embed_js_url.js").read_text()
METHOD_EMBED_JAVASCRIPT_BY_CONTENT = (root / "embed_js.js").read_text()
METHOD_EMBED_STYLE_BY_URL = (root / "embed_style_url.js").read_text()
METHOD_EMBED_STYLE_BY_CONTENT = (root / "embed_style.js").read_text()
METHOD_SELECT_VALUES = (root / "select_values.js").read_text()
METHOD_WAIT_FOR_XPATH_OR_SELECTOR = (root / "wait_for_node.js").read_text()
METHOD_ADD_PAGE_BINDING = (root / "add_binding.js").read_text()
METHOD_DELIVER_BINDING_RESULT = (root / "get_binding_result.js").read_text()
METHOD_FETCH_REQUEST = (root / "fetch_request.js").read_text()



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Callable, Literal

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    FETCH_AUTH_REQD,
    FETCH_CONTINUE,
    FETCH_CONTINUE_AUTH,
    FETCH_DISABLED,
    FETCH_ENABLED,
    FETCH_REQUEST_PAUSED,
    NETWORK_LOADING_FAILED,
    NETWORK_LOADING_FINISHED,
    NETWORK_MGR_REQUEST,
    NETWORK_MGR_REQUEST_FAILED,
    NETWORK_MGR_REQUEST_FINISHED,
    NETWORK_MGR_REQUEST_FROM_CACHE,
    NETWORK_MGR_RESPONSE,
    NETWORK_REQUEST_SERVED_FROM_CACHE,
    NETWORK_REQUEST_WILL_BE_SENT,
    NETWORK_RESPONSE_RECVD,
    NETWORK_RESPONSE_RECVD_EXTRA,

)
from mokr.frame import FrameManager
from mokr.network.event import NetworkEventManager
from mokr.network.manager.base import NetworkManager
from mokr.network.request import Request
from mokr.network.response import Response

if TYPE_CHECKING:
    from mokr.browser.page import Page


LOGGER = logging.getLogger(__name__)


class ChromeNetworkManager(NetworkManager):
    def __init__(
        self,
        page: Page,
        client: DevtoolsConnection,
        frame_manager: FrameManager,
        interception_callback_chain: list[Callable],
    ) -> None:
        """
        Class to handle requests in a given `mokr.browser.Page`.

        This class relies on network events to be emitted, and stores events
        and even requests to handle the various possible chains of events.

        The `interception_callback_chain` this class receives from the parent
        `mokr.browser.Page` is shared with spawned `mokr.network.Request`
        objects. While the request interception callback chain is run from the
        `Request` object, it is initiated during select events from this class.

        There are four possible orders that requests will be triggered and thus
        handled by `NetworkManager`:

        -  `_on_request_will_be_sent`.
        -  `_on_request_will_be_sent`, `_on_request_paused`.
        -  `_on_request_paused`, `_on_request_will_be_sent`.
        -  `_on_request_paused`, `_on_request_will_be_sent`,
            `_on_request_paused`, `_on_request_will_be_sent`,
            `_on_request_paused`, `_on_request_paused` (see
            crbug.com/1196004).

        There are a few different ways requests need to be handled due to this:
        - For `_on_request` we need the event from `_on_request_will_be_sent`
        and optionally the `fetch_request_id` (event["requestId"] also known as
        "interceptionId") from `_on_request_paused`.
        - If request interception is disabled, call `_on_request` once per call
        to `_on_request_will_be_sent`.
        - If request interception is enabled, call `_on_request` once per call
        to `_on_request_paused` (once per `fetch_request_id`).
        - Events are stored via the `mokr.network.NetworkEventManager`
        to allow for subsequent events to call `_on_request`.
        - Chains of redirect requests have the same `requestId` as
        the original request. This means events can be received in multiple
        orders for different requests in the same redirect chain such as:
            - `_on_request_will_be_sent`, `_on_request_will_be_sent`, ...
            - `_on_request_will_be_sent`, `_on_request_paused`,
                `_on_request_will_be_sent`, `_on_request_paused`, ...
            - `_on_request_will_be_sent`, `_on_request_paused`,
                `_on_request_paused`, `_on_request_will_be_sent`, ...
            - `_on_request_paused`, `_on_request_will_be_sent`,
                `_on_request_paused`, `_on_request_will_be_sent`,
                `_on_request_paused`, `_on_request_will_be_sent`,
                `_on_request_paused`, `_on_request_paused`, ...
                (see crbug.com/1196004)

        Args:
            page (Page): Parent `mokr.browser.Page`.
            client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`
                spawned by the parent `mokr.browser.Page`.
            frame_manager (FrameManager): The `mokr.frame.FrameManager`
                from the `mokr.browser.Page` that spawned this element.
            interception_callback_chain (list[Callable]): A list of callbacks
                to be passed into new `mokr.network.Request`s that will be run
                during "request" event interception, by that object. Inherited
                from the parent `mokr.browser.Page`.
        """
        self._events_to_methods = {
            FETCH_REQUEST_PAUSED: (
                lambda event: self._client._loop.create_task(
                    self._on_request_paused(event)
                )
            ),
            FETCH_AUTH_REQD: self._on_auth_required,
            NETWORK_REQUEST_WILL_BE_SENT: (
                lambda event: self._client._loop.create_task(
                    self._on_request_will_be_sent(event)
                )
            ),
            NETWORK_REQUEST_SERVED_FROM_CACHE: self._on_request_from_cache,
            NETWORK_RESPONSE_RECVD: self._on_response_received,
            NETWORK_LOADING_FINISHED: self._on_loading_finished,
            NETWORK_LOADING_FAILED: self._on_loading_failed,
            NETWORK_RESPONSE_RECVD_EXTRA: (
                lambda event: self._client._loop.create_task(
                    self._on_response_received_extra_info(event)
                )
            ),
        }
        super().__init__(
            page,
            client,
            frame_manager,
            interception_callback_chain,
        )
        self._network_event_manager = NetworkEventManager()

    @staticmethod
    def _patch_request_event_headers(
        request_will_be_sent_event: dict,
        request_paused_event: dict,
    ) -> None:
        request_will_be_sent_event['request']['headers'].update(
            request_paused_event['request']['headers']
        )

    @property
    def extra_http_headers(self) -> dict[str, str]:
        """
        Any extra HTTP headers assigned to this class. These are sent
        with every request this manager handles.
        """
        return self._extra_http_headers

    def _on_auth_required(self, event: dict) -> None:
        response = 'Default'
        if event['requestId'] in self._attempted_authentications:
            response = 'CancelAuth'
        elif self.credentials:
            self._attempted_authentications.add(event['requestId'])
            response = 'ProvideCredentials'
            params = {"response": response}
            creds = getattr(self, 'credentials', {})
            params["username"] = creds.get('username')
            params["password"] = creds.get('password')
            params = {k: v for k, v in params.items() if v}
        self._client.send(
            FETCH_CONTINUE_AUTH,
            {
                'requestId': event['requestId'],
                'authChallengeResponse': params,
            },
        )

    def _on_request_from_cache(self, event: dict) -> None:
        request = self._network_event_manager.get_request(event['requestId'])
        if request:
            request._from_memory_cache = True
        self.emit(NETWORK_MGR_REQUEST_FROM_CACHE, request)

    def _handle_request_redirect(
        self,
        request: Request,
        response_payload: dict[str, Any],
        extra_info: dict = None,
    ):
        response = Response(
            self._client,
            request,
            response_payload.get('status', 0),
            response_payload.get('headers', {}),
            response_payload.get('fromDiskCache'),
            response_payload.get('fromServiceWorker'),
            False,
            response_payload.get('securityDetails'),
            extra_info,
        )
        request._response = response
        if request not in request._redirect_chain:
            request._redirect_chain.append(request)
        self.forget_request(request, False)
        self.emit(NETWORK_MGR_RESPONSE, response)
        self.emit(NETWORK_MGR_REQUEST_FINISHED, request)

    def _emit_response_event(
        self,
        response_received: dict,
        extra_info: dict = None
    ) -> None:
        request_id = response_received['requestId']
        request = self._network_event_manager.get_request(request_id)
        # FileUpload sends a response without a matching request.
        if not request:
            return
        extraInfos = self._network_event_manager.response_extra_info(request_id)
        if extraInfos:
            LOGGER.error(
                'Unexpected extraInfo events for request'
                f' {request_id}'
            )
        # Chrome sends wrong extraInfo events for responses served from cache.
        # See https://github.com/puppeteer/puppeteer/issues/9965 and
        # https://crbug.com/1340398.
        if response_received['response']['fromDiskCache']:
            extra_info = None
        _response = response_received.get('response', {})
        response = Response(
            self._client,
            request,
            _response.get('status', 0),
            _response.get('headers', {}),
            _response.get('fromDiskCache'),
            _response.get('fromServiceWorker'),
            False,
            _response.get('securityDetails'),
            extra_info,
        )
        request._response = response
        self.emit(NETWORK_MGR_RESPONSE, response)

    def _on_response_received(self, event: dict) -> None:
        request_id = event['requestId']
        request = self._network_event_manager.get_request(request_id)
        extra_info = None
        if (
            request and not request._from_memory_cache and event['hasExtraInfo']
        ):
            _responseExtraInfo = (
                self._network_event_manager.response_extra_info(request_id)
            )
            if _responseExtraInfo:
                extra_info = _responseExtraInfo[0]
                _responseExtraInfo.remove(extra_info)
        if not extra_info:
            # Wait until we get the corresponding ExtraInfo event.
            self._network_event_manager.queue_event_group(
                request_id,
                {'responseReceivedEvent': event},
            )
            return
        self._emit_response_event(event, extra_info)

    def _on_loading_finished(self, event: dict) -> None:
        # If the response event for this request is still waiting on a
        # corresponding ExtraInfo event, then wait to emit this event too.
        queued_events = self._network_event_manager.get_queued_event_group(
            event['requestId']
        )
        if queued_events:
            queued_events['loadingFinishedEvent'] = event
        else:
            self._emit_loading_finished(event)

    def _emit_loading_finished(self, event: dict) -> None:
        request = self._network_event_manager.get_request(event['requestId'])
        # For certain requestIds we never receive requestWillBeSent event.
        # @see https://crbug.com/750469
        # Under certain conditions we never get the Network.responseReceived
        # event from protocol. @see https://crbug.com/883475
        if not request:
            return
        self.forget_request(request, True)
        self.emit(NETWORK_MGR_REQUEST_FINISHED, request)

    def _on_loading_failed(self, event: dict) -> None:
        # If the response event for this request is still waiting on a
        # corresponding ExtraInfo event, then wait to emit this event too.
        queued_events = self._network_event_manager.get_queued_event_group(
            event['requestId']
        )
        if queued_events:
            queued_events['loadingFailedEvent'] = event
        else:
            self._emit_loading_failed(event)

    def _emit_loading_failed(self, event):
        request = self._network_event_manager.get_request(event['requestId'])
        # // For certain requestIds we never receive requestWillBeSent event.
        # // @see https://crbug.com/750469
        if not request:
            return
        request._failure_text = event['errorText']
        self.forget_request(request, True)
        self.emit(NETWORK_MGR_REQUEST_FAILED, request)

    async def _apply_protocol_request_interception(self) -> None:
        if self._user_cache_disabled is None:
            self._user_cache_disabled = False
        if self._protocol_request_interception_enabled:
            await asyncio.gather(
                self._apply_protocol_cache_disabled(),
                self._client.send(
                    FETCH_ENABLED,
                    {
                        'handleAuthRequests': True,
                        'patterns': [{'urlPattern': '*'}],
                    }
                ),
            )
        else:
            await asyncio.gather(
                self._apply_protocol_cache_disabled(),
                self._client.send(FETCH_DISABLED),
            )

    async def _on_request_will_be_sent(self, event: dict) -> None:
        # Request interception doesn't happen for data URLs.
        if (
            self._user_request_interception_enabled
            and not event['request']['url'].startswith('data:')
        ):
            network_request_id = event['requestId']
            self._network_event_manager.store_request_will_be_sent(
                network_request_id,
                event,
            )
            # Fetch.requestPaused may have been sent already. Check for it.
            request_paused_event = (
                self._network_event_manager.get_request_paused(
                    network_request_id
                )
            )
            if request_paused_event:
                fetch_request_id = request_paused_event['requestId']
                self._patch_request_event_headers(event, request_paused_event)
                await self._on_request(event, fetch_request_id)
                self._network_event_manager.forget_request_paused(
                    network_request_id
                )
            return
        await self._on_request(event, None)

    async def _on_request_paused(self, event: dict) -> None:
        # CDP may send a "Fetch.requestPaused" event without or before a
        # "Network.requestWillBeSent" event.
        # CDP may also send multiple "Fetch.requestPaused" events
        # for the same "Network.requestWillBeSent" event.
        if (
            not self._user_request_interception_enabled
            and self._protocol_request_interception_enabled
        ):
            self._client.send(
                FETCH_CONTINUE,
                {'requestId': event['requestId']},
            )
        network_request_id = event.get('networkId')
        fetch_request_id = event['requestId']
        if not network_request_id:
            await self._on_request_without_network_instrumentation(event)
            return

        request_will_be_sent_event = (
            self._network_event_manager.get_request_will_be_sent(
                network_request_id
            )
        )
        # Redirect requests have the same requestId.
        if request_will_be_sent_event and (
            (
                request_will_be_sent_event['request']['url']
                != event['request']['url']
            )
            or (
                request_will_be_sent_event['request']['method']
                != event['request']['method']
            )
        ):
            self._network_event_manager.forget_request_will_be_sent(
                network_request_id
            )
            request_will_be_sent_event = None
        if request_will_be_sent_event:
            self._patch_request_event_headers(request_will_be_sent_event, event)
            await self._on_request(request_will_be_sent_event, fetch_request_id)
        else:
            self._network_event_manager.store_request_paused(
                network_request_id,
                event,
            )

    async def _on_request_without_network_instrumentation(
        self,
        event: dict,
    ) -> None:
        mokr_request_uuid = self._inspect_stack_for_mokr_uuid(event)
        # If an event has no networkId it should not have any network events.
        # Still want to dispatch it for the interception by the user.
        frame = (
            self._frame_manager.frame(event['frameId'])
            if event.get('frameId') else None
        )
        request = Request(
            self._page,
            self._client,
            event['requestId'],
            None,
            False,
            self._user_request_interception_enabled,
            event.get('request', {}).get('url'),
            event.get('type', ''),
            event.get('request', {}),
            frame,
            [],
            self._interception_callback_chain,
        )
        # This should never happen here.
        if mokr_request_uuid:
            request._mokr_request_uuid = mokr_request_uuid
        self.emit(NETWORK_MGR_REQUEST, request)
        await request._finalize_interceptions()

    async def _on_request(
        self,
        event: dict,
        fetch_request_id: str = None,
    ) -> None:
        redirect_chain = []
        mokr_request_uuid = self._inspect_stack_for_mokr_uuid(event)
        if event.get('redirectResponse'):
            # Want to emit "response" and "requestfinished" events for the
            # redirectResponse, but can't do so unless we have an event's
            # "responseExtraInfo" ready to pair it up with. If we don't have any
            # "responseExtraInfo" data saved in our queue under this event's
            # requestId, must wait for the next one to emit both "response" and
            # "requestfinished" events, but we also should wait to emit this
            # `mokr.network.Request` via a "NetworkManager.Request" event
            # because it should come after "response"/"requestfinished" events.
            redirect_response_extra_info = None
            if event.get('redirectHasExtraInfo'):
                _response_extra_info = (
                    self._network_event_manager.response_extra_info(
                        event['requestId']
                    )
                )
                if _response_extra_info:
                    redirect_response_extra_info = _response_extra_info[0]
                    _response_extra_info.remove(redirect_response_extra_info)
                if not redirect_response_extra_info:
                    self._network_event_manager.queue_redirect_info(
                        event['requestId'],
                        {"event": event, "fetchRequestId": fetch_request_id},
                    )
                    return
            request = self._network_event_manager.get_request(
                event['requestId']
            )
            # If we connect late, may have missed requestWillBeSent.
            if request:
                self._handle_request_redirect(
                    request,
                    event.get('redirectResponse'),
                    redirect_response_extra_info
                )
                redirect_chain = request._redirect_chain
        frame = (
            self._frame_manager.frame(event['frameId'])
            if event.get('frameId') else None
        )
        is_navigation_request = bool(
            event.get('requestId') == event.get('loaderId')
            and event.get('type') == 'Document'
        )
        # Bind the same FetchDomain request uuid to redirect requests.
        if (
            not mokr_request_uuid
            and any(r._mokr_request_uuid for r in redirect_chain)
        ):
            mokr_request_uuid = next(
                iter(r._mokr_request_uuid for r in redirect_chain)
            )
        request = Request(
            self._page,
            self._client,
            event['requestId'],
            fetch_request_id,
            is_navigation_request,
            self._user_request_interception_enabled,
            event.get('request', {}).get('url'),
            event.get('type', ''),
            event.get('request', {}),
            frame,
            redirect_chain,
            self._interception_callback_chain,
        )
        # Append final request to redirect chain.
        if request not in request._redirect_chain:
            request._redirect_chain.append(request)
        if mokr_request_uuid:
            request._mokr_request_uuid = mokr_request_uuid
        self._network_event_manager.store_request(event['requestId'], request)
        self.emit(NETWORK_MGR_REQUEST, request)
        await request._finalize_interceptions()

    async def _on_response_received_extra_info(self, event: dict) -> None:
        # We may have skipped a redirect response/request pair due to waiting
        # for this ExtraInfo event.
        request_id = event['requestId']
        redirect_info = self._network_event_manager.take_queued_redirect_info(
            request_id
        )
        if redirect_info:
            self._network_event_manager.response_extra_info(
                request_id
            ).append(event)
            await self._on_request(
                redirect_info['event'],
                redirect_info['fetchRequestId'],
            )
            return
        # We may have skipped response and loading events because we didn't have
        # this ExtraInfo event yet. If so, emit those events now.
        queued_events = self._network_event_manager.get_queued_event_group(
            request_id
        )
        if queued_events:
            self._network_event_manager.forget_queued_event_group(request_id)
            self._emit_response_event(
                queued_events['responseReceivedEvent'],
                event,
            )
            if queued_events.get('loadingFinishedEvent'):
                self._emit_loading_finished(
                    queued_events['loadingFinishedEvent']
                )
            if queued_events.get('loadingFailedEvent'):
                self._emit_loading_failed(queued_events['loadingFailedEvent'])
            return
        # Wait until we get another event that can use this ExtraInfo event.
        self._network_event_manager.response_extra_info(
            request_id
        ).append(event)

    def forget_request(self, request: Request, events: bool) -> None:
        """
        Drop a request from the `mokr.network.NetworkEventManager` and
        remove from attempted authentications listing, if relevant.

        Optionally forget all events tied to the remote request networkID, too.

        Args:
            request (Request): `mokr.request.Request` to forget.
            events (bool): If True, drop all events stored for the request, too.
        """
        request_id = request._request_id
        interception_id = request._interception_id
        self._network_event_manager.forget_request(request_id)
        if interception_id:
            try:
                self._attempted_authentications.remove(interception_id)
            except KeyError:
                pass
        if events:
            self._network_event_manager.forget(request_id)

    def get_in_flight_requests_count(self) -> int:
        """
        Get the number of active requests (not resolved).
        Wraps `mokr.network.NetworkEventManager.get_in_flight_requests_count`.

        Returns:
            int: Number of active requests.
        """
        return self._network_event_manager.get_in_flight_requests_count()

    async def set_credentials(
        self,
        credentials: dict[Literal["username", "password"], str],
    ) -> None:
        """
        Set the credentials to use for HTTP authentication.

        Args:
            credentials (dict[str, str]): A dictionary with credentials,
                keyed as "username" and "password".
        """
        self.credentials = credentials
        enabled = (
            self._user_request_interception_enabled or not self.credentials
        )
        if enabled == self._protocol_request_interception_enabled:
            return
        self._protocol_request_interception_enabled = enabled
        self._apply_protocol_request_interception()

    async def set_extra_http_headers(
        self,
        extra_http_headers: dict[str, str],
    ) -> None:
        """
        Set extra headers to be sent with every request.

        Args:
            extra_http_headers (dict[str, str]): A dictionary of headers.

        Raises:
            ValueError: Raised if any header value is not string.
        """
        for key, value in extra_http_headers.items():
            if not isinstance(value, str):
                raise ValueError(
                    f'Expected value of header "{key}"'
                    f' to be str, not {type(value)}.'
                )
        self._extra_http_headers.update(extra_http_headers)
        await self._apply_extra_http_headers()

    async def set_request_interception(self, choice: bool) -> None:
        """
        Disable or enabled request interception (enabled by default).
        See `mokr.browser.Page.on` for details on using the request interception
        callback chain.

        Args:
            choice (bool): False to disable, True to enable.
        """
        self._user_request_interception_enabled = choice
        enabled = (
            self._user_request_interception_enabled or bool(self.credentials)
        )
        if enabled == self._protocol_request_interception_enabled:
            return
        self._protocol_request_interception_enabled = enabled
        await self._apply_protocol_request_interception()
